local envHasGetChildren = Inspect.System.Version ~= nil

local lib = {
  debug = false,
  recycled = {},
}

if not envHasGetChildren then
  lib.created = {}
end

local function FindContext(frame)
  local parent = frame:GetParent()
  if parent == nil then
    return frame
  else
    return FindContext(parent)
  end
end

function lib.CreateFrame(frameType, namePrefix, parent)
  local context = FindContext(parent)
  if lib.recycled[context] == nil then
    lib.recycled[context] = {}
    if not envHasGetChildren then
      lib.created[context] = {}
    end
  end
  local recycled = lib.recycled[context]

  if recycled[frameType] == nil then
    recycled[frameType] = {}
  end
  local frames = recycled[frameType]

  local frame = table.remove(frames)
  if frame == nil then
    frame = UI.CreateFrame(frameType, namePrefix..tostring(#frames), parent)
    frame.frameType = frameType
    if lib.debug then
      print("Created new "..frameType.." frame: "..frame:GetName())
    end
  else
    frame:SetParent(parent)
    if lib.debug then
      print("Used recycled frame: "..frame:GetName())
    end
  end

  frame:SetVisible(true)

  if not envHasGetChildren then
    local created = lib.created[context]
    table.insert(created, frame)
  end

  return frame
end

function lib.RecycleFrame(frame)
  local context = FindContext(frame)
  if lib.recycled[context] == nil then
    lib.recycled[context] = {}
    if not envHasGetChildren then
      lib.created[context] = {}
    end
  end
  local recycled = lib.recycled[context]

  local children

  if not envHasGetChildren then
    children = {}

    local frameIndex

    local created = lib.created[context]
    for i, v in ipairs(created) do
      if v == frame then
        -- can't remove immediately since the next iteration would skip a frame
        frameIndex = i
      elseif v:GetParent() == frame then
        table.insert(children, v)
      end
    end

    if frameIndex ~= nil then
      table.remove(created, frameIndex)
    end
  else
    children = frame:GetChildren()
  end

  if recycled[frame.frameType] == nil then
    recycled[frame.frameType] = {}
  end

  table.insert(recycled[frame.frameType], frame)
  frame:SetVisible(false)
  frame:SetParent(context)
  frame:ClearAll()
  frame:ClearWidth()
  frame:ClearHeight()
  if frame.SetText then
    frame:SetText("")
  end
  if lib.debug then
    print("Recycled frame: "..frame:GetName())
  end

  for i, v in ipairs(children) do
    lib.RecycleFrame(v)
  end
end

function lib.Test()
  local c1 = UI.CreateContext("context1"); local f1 = Library.LibFrameRecycler.CreateFrame("Frame", "context1frame", c1)
  dump({f1, f1:GetName()})
  Library.LibFrameRecycler.RecycleFrame(f1)
  local f2 = Library.LibFrameRecycler.CreateFrame("Frame", "context1frame", c1)
  dump({f2, f2:GetName()})
  Library.LibFrameRecycler.RecycleFrame(f2)
  local c2 = UI.CreateContext("context2")
  local f3 = Library.LibFrameRecycler.CreateFrame("Frame", "context2frame", c2)
  dump({f3, f3:GetName()})
end

-- Make the library globally accessible
Library = Library or {}
Library.LibFrameRecycler = lib
